import operator
from collections import defaultdict

from packaging.utils import canonicalize_name
from packaging.version import parse as parse_version
from wheel.install import WheelFile as _WheelFile
from wheel.install import BadWheelFile


class WheelFile:

    def __init__(self, filename):
        self.filename = filename

        whl = _WheelFile(self.filename)
        self._info = whl.parsed_filename.groupdict()

    @property
    def name(self):
        return canonicalize_name(self._info['name'])

    @property
    def version(self):
        return parse_version(self._info['ver'])


class WheelSet:
    def __init__(self):
        self._wheels = defaultdict(list)

    def add(self, filename):
        try:
            whl = WheelFile(filename)
        except BadWheelFile:
            return False
        else:
            self._wheels[whl.name].append(whl)
        return True

    def find(self, requirement):
        name = canonicalize_name(requirement.name)

        candidates = self._wheels[name]
        if not candidates:
            return

        candidates.sort(key=operator.attrgetter('version'), reverse=True)
        for candidate in candidates:
            if requirement.specifier.contains(candidate.version):
                return candidate


def collect_filenames(filenames, requirements):
    wheelset = WheelSet()
    for filename in filenames:
        wheelset.add(filename)

    result = []
    missing = []
    for requirement in requirements:
        package = wheelset.find(requirement)
        if package:
            result.append(package.filename)
        else:
            missing.append(requirement)
    return result, missing



from spaken import finder
from spaken.helpers import Requirement


def test_find():
    packages = [
        'django_healthchecks-1.4.0-py2.py3-none-any.whl',
        'django-healthchecks-1.3.0-py2.py3-none-any.whl',
        'django-healthchecks-1.3.1-py2.py3-none-any.whl',
    ]

    wheel_set = finder.WheelSet()
    for name in packages:
        wheel_set.add(name)

    whl = wheel_set.find(Requirement('django-healthchecks==1.4.0'))
    assert whl.filename == 'django_healthchecks-1.4.0-py2.py3-none-any.whl'

    whl = wheel_set.find(Requirement('django-healthchecks<1.4.0'))
    assert whl.filename == 'django-healthchecks-1.3.1-py2.py3-none-any.whl'

    whl = wheel_set.find(Requirement('django-healthchecks>=1.3.0,<1.4.0'))
    assert whl.filename == 'django-healthchecks-1.3.1-py2.py3-none-any.whl'

    whl = wheel_set.find(Requirement('django-healthchecks'))
    assert whl.filename == 'django_healthchecks-1.4.0-py2.py3-none-any.whl'


def test_invalid_file():
    packages = [
        'django_healthchecks-1.4.0-py2.py3-none-any.invalid',
    ]

    wheel_set = finder.WheelSet()
    for name in packages:
        wheel_set.add(name)

    whl = wheel_set.find(Requirement('django-healthchecks==1.4.0'))
    assert whl is None


def test_collect_filenames():
    filenames = [
        'django_healthchecks-1.4.0-py2.py3-none-any.invalid',
        'django_healthchecks-1.4.0-py2.py3-none-any.whl',
        'django-healthchecks-1.3.0-py2.py3-none-any.whl',
        'django-healthchecks-1.3.1-py2.py3-none-any.whl',
    ]

    requirements = [
        Requirement('Django-healthchecks==1.4.0'),
        Requirement('wsgi-basic-auth-healthchecks==1.1.0'),
    ]

    result, missing = finder.collect_filenames(filenames, requirements)
    assert result == ['django_healthchecks-1.4.0-py2.py3-none-any.whl']
    assert missing == [Requirement('wsgi-basic-auth-healthchecks==1.1.0')]



import os
from urllib.parse import urlparse
from botocore.exceptions import BotoCoreError

import boto3

from spaken.exceptions import StorageException


class S3Storage:

    def __init__(self, uri):
        bucket_name, bucket_path = parse_bucket_uri(uri)
        s3 = boto3.resource('s3')
        try:
            s3.meta.client.head_bucket(Bucket=bucket_name)
        except BotoCoreError:
            raise StorageException

        self._bucket = s3.Bucket(bucket_name)
        self._path = bucket_path

    def list_files(self):
        if self._path:
            objects = self._bucket.objects.filter(Prefix=self._path)
        else:
            objects = self._bucket.objects.all()

        return [os.path.relpath(obj.key, self._path) for obj in objects]

    def upload(self, source, filename):
        key = self._path + filename
        with open(source, 'rb') as fh:
            self._bucket.put_object(Key=key, Body=fh.read())

    def get(self, filename, destination):
        key = self._path + filename
        self._bucket.download_file(key, destination)


def parse_bucket_uri(uri):
    result = urlparse(uri)

    path = result.path
    if path:
        path = path[1:]
    if not path.endswith('/'):
        path += '/'

    # Remove leading /
    path = path.lstrip('/')
    return result.netloc, path



from setuptools import find_packages, setup


with open('README.rst', 'r') as fh:
    description = '\n'.join(fh.readlines())


tests_require = [
    'moto==1.3.3',
    'pytest>=3.5.1',
    'pytest-cov>=2.5.1',
]


setup(
    name='spaken',
    version='0.0.9',
    description=description,
    url='https://github.com/labd/spaken',
    author="Michael van Tellingen",
    author_email="michaelvantellingen@gmail.com",
    install_requires=[
        'boto3',
        'click>=6.7',
        'pip>=10.0',
        'packaging>=17.0',
    ],
    tests_require=tests_require,
    extras_require={'test': tests_require},
    package_dir={'': 'src'},
    packages=find_packages('src'),
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
    ],
    entry_points={
        'console_scripts': {
            'spaken = spaken.main:main'
        }
    },
    zip_safe=False,
)
